"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.defaultConfig = exports.Preszr = void 0;
const utils_1 = require("./utils");
const data_1 = require("./data");
const encodings_1 = require("./encodings");
const encode_context_1 = require("./encode/encode-context");
const fixed_indexes_1 = require("./encodings/fixed-indexes");
const decode_context_1 = require("./encode/decode-context");
const texts_1 = require("./errors/texts");
/**
 * The class used to encode and decode things in the preszr format.
 */
class Preszr {
    constructor(config) {
        var _a;
        this._store = (0, encodings_1.getDefaultStore)();
        if (config && typeof config !== "object") {
            throw (0, texts_1.getErrorByCode)("config/bad-type")(config);
        }
        config !== null && config !== void 0 ? config : (config = exports.defaultConfig);
        if (config.encodes && !Array.isArray(config.encodes)) {
            throw (0, texts_1.getErrorByCode)("config/encodes/not-array")(config.encodes);
        }
        this._store.add(...((_a = config.encodes) !== null && _a !== void 0 ? _a : []));
    }
    _checkInputHeader(input) {
        let versionInfo = "";
        if (!Array.isArray(input)) {
            throw (0, texts_1.getErrorByCode)("decode/input/not-array")(input);
        }
        else {
            const header = input === null || input === void 0 ? void 0 : input[0];
            if (!header) {
                throw (0, texts_1.getErrorByCode)("decode/input/empty-array")();
            }
            else if (!Array.isArray(header)) {
                throw (0, texts_1.getErrorByCode)("decode/input/bad-header")(header);
            }
            else {
                versionInfo = header[0];
                if (versionInfo == null) {
                    throw (0, texts_1.getErrorByCode)("decode/input/header/empty-array")();
                }
                else if (typeof versionInfo !== "string") {
                    throw (0, texts_1.getErrorByCode)("decode/input/version/not-string")(versionInfo);
                }
                else if (!(0, utils_1.isNumeric)(versionInfo)) {
                    throw (0, texts_1.getErrorByCode)("decode/input/version/not-numeric")(versionInfo);
                }
                else if (versionInfo !== utils_1.version) {
                    throw (0, texts_1.getErrorByCode)("decode/input/version/mismatch")(versionInfo, utils_1.version);
                }
                else if (!Array.isArray(header[1])) {
                    throw (0, texts_1.getErrorByCode)("decode/input/header/no-keys")();
                }
                else if (typeof header[2] !== "object" ||
                    !header[2] ||
                    Array.isArray(header[2])) {
                    throw (0, texts_1.getErrorByCode)("decode/input/header/no-map")();
                }
                else if (typeof header[3] !== "object" ||
                    !header[3] ||
                    Array.isArray(header[2])) {
                    throw (0, texts_1.getErrorByCode)("decode/input/header/no-metadata")();
                }
                else if (header.length > 4) {
                    throw (0, texts_1.getErrorByCode)("decode/input/header/too-long")(header);
                }
                else if (input.length === 1) {
                    throw (0, texts_1.getErrorByCode)("decode/input/no-data")();
                }
            }
        }
    }
    decode(input) {
        if (typeof input !== "object" || input === null) {
            const tryScalar = (0, data_1.tryDecodeScalar)(input);
            switch (tryScalar) {
                case data_1.noResultPlaceholder:
                    break;
                case data_1.badType:
                    throw (0, texts_1.getErrorByCode)("decode/input/bad-type")(input);
                case data_1.unknownScalar:
                    throw (0, texts_1.getErrorByCode)("decode/input/unknown-scalar")(input);
                default:
                    return tryScalar;
            }
        }
        input = input;
        // We check the header is in Preszr format.
        this._checkInputHeader(input);
        // Deconstruct the header to its parts
        const [[, encodingKeys, encodingSpec, metadata]] = input;
        const targetArray = Array(input.length - 1);
        const needToInit = new Map();
        // For optimization purposes, create one instance of `InitContext`.
        const ctx = new decode_context_1.DecodeContext(targetArray);
        const encodingByIndex = [];
        // Check all encoding keys are present.
        for (const encodingKey of encodingKeys) {
            encodingByIndex.push(this._store.mustGetByKey(encodingKey));
        }
        // Start decoding the payload.
        for (let i = 1; i < input.length; i++) {
            const encodingIndex = encodingSpec[i];
            const cur = input[i];
            let encoding;
            if (encodingIndex == null) {
                if (typeof cur === "string") {
                    targetArray[i] = cur;
                    continue;
                }
                if (Array.isArray(cur)) {
                    encoding = encodings_1.arrayEncoding;
                }
                else {
                    encoding = encodings_1.objectEncoding;
                }
            }
            else if (encodingIndex === fixed_indexes_1.FixedIndexes.UnknownSymbol) {
                targetArray[i] = (0, utils_1.getUnrecognizedSymbol)(metadata[i]);
                continue;
            }
            else if (encodingIndex < fixed_indexes_1.FixedIndexes.End) {
                encoding = this._store.mustGetByIndex(encodingIndex);
            }
            else {
                encoding = encodingByIndex[encodingIndex - fixed_indexes_1.FixedIndexes.End];
            }
            if (typeof encoding.encodes === "symbol") {
                targetArray[i] = encoding.encodes;
                continue;
            }
            const protoEnc = encoding;
            ctx.metadata = metadata[i];
            ctx.self = protoEnc;
            targetArray[i] = protoEnc.decoder.create(cur, ctx);
            ctx.self = null;
            if (protoEnc.decoder.init) {
                needToInit.set(i, protoEnc);
            }
        }
        ctx.next();
        for (const key of needToInit.keys()) {
            const encoding = needToInit.get(key);
            ctx.metadata = metadata[key];
            ctx.self = encoding;
            encoding.decoder.init(targetArray[key], input[key], ctx);
            ctx.self = null;
        }
        return targetArray[1];
    }
    encode(root) {
        const tryScalar = (0, data_1.tryEncodeScalar)(root);
        if (tryScalar !== data_1.noResultPlaceholder) {
            return tryScalar;
        }
        const ctx = new encode_context_1.EncodeCtx(this._store);
        ctx.encode(root);
        const result = ctx.finish();
        return result;
    }
}
exports.Preszr = Preszr;
exports.defaultConfig = {
    encodes: []
};
//# sourceMappingURL=core.js.map