"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.EncodingStore = void 0;
const interface_1 = require("../interface");
const utils_1 = require("../encodings/utils");
const utils_2 = require("../utils");
const fixed_indexes_1 = require("../encodings/fixed-indexes");
const encodings_1 = require("../encodings");
const user_encoding_1 = require("../encodings/user-encoding");
const texts_1 = require("../errors/texts");
class EncodingStore {
    constructor() {
        // We use multiple caches to speed up encoding and decoding.
        // It's a space/time trade-off, but it'll barely take any space.
        // This maintains a full list of all versions of all prototype encodings.
        // it's used for various checks and to built the caches.
        this._nameToProtoEncodings = new Map();
        // This maintains all symbol encodings and is also used when encoding.
        this._symbolToEncoding = new Map();
        // Matches a prototype to an encoding. Used for some checks.
        this._protoToEncoding = new Map();
        // A cache that quickly matches an encoding key, e.g. Custom/whatever.v5 -> encoding.
        // used when decoding.
        // Both symbol and proto encodings use this cache, but their keys are different.
        // Proto encodings have keys ending with .v${version}, and symbol encodings have keys
        // ending with .S
        this._keyToEncoding = new Map();
        // A "map" of index to built-in encoding. Has empty indexes.
        this._indexToEncoding = Array(fixed_indexes_1.FixedIndexes.End);
    }
    all() {
        return [...this.getProtoEncodings(), ...this.getSymbolEncodings()];
    }
    _getBuiltInEncoding(prototype) {
        const existing = this._protoToEncoding.get(prototype);
        if (!existing) {
            return null;
        }
        return this._isBuiltIn(existing) ? existing : null;
    }
    _isBuiltIn(encoding) {
        const fixedIndex = encoding.fixedIndex;
        return fixedIndex != null && fixedIndex < fixed_indexes_1.FixedIndexes.End;
    }
    _registerProtos(encoding) {
        const existingEncoding = this._protoToEncoding.get(encoding.encodes);
        // It's illegal to have two different encodings for the same prototype, since it becomes ambiguous.
        if (existingEncoding && existingEncoding.name !== encoding.name) {
            throw (0, texts_1.getErrorByCode)("config/proto/proto-collision")(existingEncoding, encoding);
        }
        // However, it's okay to have the same encoding with different versions reference different prototypes.
        if (!existingEncoding || existingEncoding.version < encoding.version) {
            this._protoToEncoding.set(encoding.encodes, encoding);
        }
    }
    _addEncoding(encoding) {
        // After adding encodings, the cache must be rebuilt.
        if (encoding instanceof interface_1.SymbolEncoding) {
            this._addSymbolEncoding(encoding);
        }
        else if (encoding instanceof interface_1.PrototypeEncoding) {
            this._addProtoEncoding(encoding);
        }
        else {
            throw (0, texts_1.getErrorByCode)("bug/config/unknown-encoding")(encoding);
        }
    }
    _maybeRegisterFixedIndexEncoding(encoding) {
        const fixed = encoding.fixedIndex;
        if (fixed != null) {
            const existing = this._indexToEncoding[fixed];
            if (existing && existing.name !== encoding.name) {
                throw (0, texts_1.getErrorByCode)("bug/config/fixed-index-collision")(encoding, existing);
            }
            this._indexToEncoding[fixed] = encoding;
        }
    }
    makeEncodingFromCtor(ctor) {
        return this.makeEncodingFromProtoSpec({
            encodes: ctor
        });
    }
    makeEncodingFromProtoSpec(specifier) {
        let proto;
        if (typeof specifier.encodes === "function") {
            proto = (0, utils_2.getProto)(specifier.encodes);
            if (!proto) {
                throw (0, texts_1.getErrorByCode)("config/proto/couldnt-get-prototype")(specifier.encodes);
            }
        }
        else if (typeof specifier.encodes === "object") {
            proto = specifier.encodes;
        }
        else {
            throw (0, texts_1.getErrorByCode)("config/spec/bad-encodes")(specifier.encodes);
        }
        const builtIn = this._getBuiltInEncoding(proto);
        let name = specifier.name;
        let fixedIndex = undefined;
        if (builtIn) {
            if (name) {
                throw (0, texts_1.getErrorByCode)("config/spec/name-illegal-builtin")(proto);
            }
            name = builtIn.name;
            fixedIndex = builtIn.fixedIndex;
        }
        else {
            name !== null && name !== void 0 ? name : (name = (0, utils_2.getPrototypeName)(proto));
            if (!name) {
                throw (0, texts_1.getErrorByCode)("config/spec/proto/no-name")(proto);
            }
        }
        const encoding = new user_encoding_1.UserEncoding({
            encodes: proto,
            encode: specifier.encode,
            version: specifier.version,
            name,
            decoder: specifier.decoder
        }, fixedIndex);
        return encoding;
    }
    makeEncoding(encoding) {
        if (typeof encoding === "function") {
            return this.makeEncodingFromCtor(encoding);
        }
        if (typeof encoding === "symbol") {
            return interface_1.SymbolEncoding.fromSymbol(encoding);
        }
        if (typeof encoding !== "object") {
            throw (0, texts_1.getErrorByCode)("config/spec/bad-type")(encoding);
        }
        if (!encoding.encodes) {
            throw (0, texts_1.getErrorByCode)("config/spec/no-encodes")();
        }
        if ((0, utils_1.isSymbolSpecifier)(encoding)) {
            return interface_1.SymbolEncoding.fromSpecifier(encoding);
        }
        return this.makeEncodingFromProtoSpec(encoding);
    }
    add(...encodings) {
        for (const encoding of encodings) {
            const fullEncoding = encoding instanceof interface_1.BaseEncoding
                ? encoding
                : this.makeEncoding(encoding);
            this._addEncoding(fullEncoding);
        }
    }
    _addProtoEncoding(encoding) {
        const existingByProto = this._protoToEncoding.get(encoding.encodes);
        if (existingByProto) {
            encoding.mustBeCompatible(existingByProto);
        }
        let versioned = this._nameToProtoEncodings.get(encoding.name);
        if (!versioned) {
            // Initialize the versioned encoding list if it doesn't exist
            versioned = new Map();
            this._nameToProtoEncodings.set(encoding.name, versioned);
        }
        if (versioned.get(encoding.version)) {
            throw (0, texts_1.getErrorByCode)("config/proto/encoding-exists")(encoding);
        }
        versioned.set(encoding.version, encoding);
        this._registerProtos(encoding);
        // The latest encoding version is kept under -1 for easy access.
        const maxVersionEncoding = versioned.get(-1);
        if (!maxVersionEncoding ||
            maxVersionEncoding.version < encoding.version) {
            versioned.set(-1, encoding);
            this._maybeRegisterFixedIndexEncoding(encoding);
            // If the max version was set, that means the proto cache is obsolete
            this._cacheProtoToEncoding = undefined;
        }
        versioned.set(encoding.version, encoding);
        this._keyToEncoding.set(encoding.key, encoding);
        return encoding;
    }
    *getProtoEncodings() {
        for (const [, encoding] of this._protoToEncoding) {
            yield encoding;
        }
    }
    getSymbolEncodings() {
        return this._symbolToEncoding.values();
    }
    _addSymbolEncoding(encoding) {
        const existingBySymbol = this._symbolToEncoding.get(encoding.encodes);
        const existingByKey = this._keyToEncoding.get(encoding.key);
        if (existingByKey) {
            throw (0, texts_1.getErrorByCode)("config/symbol/name-exists")(existingByKey);
        }
        if (existingBySymbol) {
            throw (0, texts_1.getErrorByCode)("config/symbol/already-encoded")(existingBySymbol, encoding);
        }
        this._symbolToEncoding.set(encoding.encodes, encoding);
        this._keyToEncoding.set(encoding.key, encoding);
        this._maybeRegisterFixedIndexEncoding(encoding);
        return encoding;
    }
    mayGetBySymbol(s) {
        return this._symbolToEncoding.get(s);
    }
    mustGetByIndex(ix) {
        const encoding = this._indexToEncoding[ix];
        if (!encoding) {
            if (ix < fixed_indexes_1.FixedIndexes.End) {
                throw (0, texts_1.getErrorByCode)("decode/map/unknown-builtin-index")(ix);
            }
            throw (0, texts_1.getErrorByCode)("decode/map/unknown-index")(ix);
        }
        return encoding;
    }
    mustGetByKey(key) {
        var _a;
        const encoding = this._keyToEncoding.get(key);
        if (encoding) {
            return encoding;
        }
        // If it failed, this is going to be an error
        // which means there are no performance considerations. Let's gather as much
        // info as possible.
        const info = (0, utils_1.mustParseEncodingKey)(key);
        if (info.type === "symbol") {
            throw (0, texts_1.getErrorByCode)("decode/keys/unknown-symbol")(info.name);
        }
        else {
            const namedProtoEncoding = (_a = this._nameToProtoEncodings
                .get(info.name)) === null || _a === void 0 ? void 0 : _a.get(-1);
            if (!namedProtoEncoding) {
                throw (0, texts_1.getErrorByCode)("decode/keys/unknown-proto")(info.name);
            }
            throw (0, texts_1.getErrorByCode)("decode/keys/unknown-proto-version")(info.name, info.version, namedProtoEncoding);
        }
    }
    _makeWorkingProtoCache() {
        // We add an entry to the working encoding cache for every registered
        // encoding.
        const cache = new WeakMap();
        for (const encoding of this.getProtoEncodings()) {
            if (!encoding) {
                (0, texts_1.getErrorByCode)("bug/encode/no-highest-version")(encoding);
            }
            cache.set(encoding.encodes, encoding);
        }
        return cache;
    }
    mustGetByProto(obj) {
        var _a;
        if (!this._cacheProtoToEncoding) {
            this._cacheProtoToEncoding = this._makeWorkingProtoCache();
        }
        const fromCache = this._cacheProtoToEncoding.get(obj);
        if (fromCache) {
            return fromCache;
        }
        let foundEncoding;
        const chain = [];
        for (let proto = obj;; proto = (_a = Object.getPrototypeOf(proto)) !== null && _a !== void 0 ? _a : encodings_1.nullPlaceholder) {
            const cached = this._cacheProtoToEncoding.get(proto);
            if (cached !== undefined) {
                foundEncoding = cached;
                break;
            }
            if (proto === encodings_1.nullPlaceholder) {
                throw (0, texts_1.getErrorByCode)("bug/encode/proto-without-match")(chain[0]);
            }
            chain.push(proto);
        }
        for (const proto of chain) {
            this._cacheProtoToEncoding.set(proto, foundEncoding);
        }
        return foundEncoding;
    }
}
exports.EncodingStore = EncodingStore;
//# sourceMappingURL=store.js.map