"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.unsupportedEncodings = exports.getPrototypeEncoder = exports.getPrototypeDecoder = exports.nullPrototypeEncoding = exports.arrayEncoding = exports.objectEncoding = exports.encodeObject = exports.decodeObject = exports.nullPlaceholder = void 0;
const interface_1 = require("../interface");
const utils_1 = require("../utils");
const fixed_indexes_1 = require("./fixed-indexes");
const opt_types_1 = require("../opt-types");
const utils_2 = require("./utils");
exports.nullPlaceholder = {};
function getAllOwnKeys(obj, onlyEnumerable) {
    const keys = Reflect.ownKeys(obj);
    if (onlyEnumerable) {
        return keys.filter(x => Object.prototype.propertyIsEnumerable.call(obj, x));
    }
    return keys;
}
function decodeObject(target, input, ctx) {
    let stringKeys = input;
    if (Array.isArray(input)) {
        let symbolKeys;
        [symbolKeys, stringKeys] = input;
        for (const key of Object.keys(symbolKeys)) {
            const value = symbolKeys[key];
            target[ctx.decode(key)] = ctx.decode(value);
        }
    }
    for (const key of Object.keys(stringKeys)) {
        const value = stringKeys[key];
        target[key] = ctx.decode(value);
    }
    return target;
}
exports.decodeObject = decodeObject;
function encodeObject(input, ctx, alsoNonEnumerable, explicitlyInclude = []) {
    const strKeyObject = {};
    let symbKeyObject;
    for (const key of getAllOwnKeys(input, !alsoNonEnumerable)) {
        const value = input[key];
        if (typeof key === "symbol") {
            symbKeyObject !== null && symbKeyObject !== void 0 ? symbKeyObject : (symbKeyObject = {});
            symbKeyObject[ctx.encode(key)] = ctx.encode(value);
        }
        else {
            strKeyObject[key] = ctx.encode(value);
        }
    }
    for (const key of explicitlyInclude) {
        if (!(key in strKeyObject) && key in input) {
            strKeyObject[key] = ctx.encode(input[key]);
        }
    }
    if (symbKeyObject) {
        return [symbKeyObject, strKeyObject];
    }
    ctx._isImplicit = true;
    return strKeyObject;
}
exports.encodeObject = encodeObject;
exports.objectEncoding = (0, utils_2.defineProtoEncoding)(class ObjectEncoding extends interface_1.PrototypeEncoding {
    constructor() {
        super(...arguments);
        this.version = 0;
        this.encodes = Object.prototype;
        this.fixedIndex = fixed_indexes_1.FixedIndexes.Object;
        this.name = (0, utils_1.getBuiltInEncodingName)("object");
        this.decoder = {
            create(encodedValue, ctx) {
                return {};
            },
            init(target, encoded, ctx) {
                decodeObject(target, encoded, ctx);
            }
        };
    }
    encode(input, ctx) {
        return encodeObject(input, ctx, false);
    }
});
function encodeAsSparseArray(input, ctx) {
    // Sparse arrays are serialized like objects.
    const result = encodeObject(input, ctx, false);
    ctx._isImplicit = false;
    return result;
}
exports.arrayEncoding = (0, utils_2.defineProtoEncoding)(class ArrayEncoding extends interface_1.PrototypeEncoding {
    constructor() {
        super(...arguments);
        this.name = (0, utils_1.getBuiltInEncodingName)("array");
        this.version = 0;
        this.fixedIndex = fixed_indexes_1.FixedIndexes.Array;
        this.encodes = Array.prototype;
        this.decoder = {
            create(encodedValue, ctx) {
                return [];
            },
            init(target, input, ctx) {
                if (!Array.isArray(input)) {
                    // Decode similarly to objects
                    decodeObject(target, input, ctx);
                    return;
                }
                for (let i = 0; i < input.length; i++) {
                    target[i] = ctx.decode(input[i]);
                }
            }
        };
    }
    encode(input, ctx) {
        const keys = Object.keys(input);
        const isSparseCanFalseNegative = input.length !== keys.length;
        if (isSparseCanFalseNegative) {
            return encodeAsSparseArray(input, ctx);
        }
        // The array still might be sparse, even after that check.
        const newArray = [];
        for (let i = 0; i < keys.length; i++) {
            if (i !== +keys[i]) {
                return encodeAsSparseArray(input, ctx);
            }
            newArray.push(ctx.encode(input[i]));
        }
        ctx._isImplicit = true;
        return newArray;
    }
});
exports.nullPrototypeEncoding = (0, utils_2.defineProtoEncoding)(class NullPrototypeEncoding extends interface_1.PrototypeEncoding {
    constructor() {
        super(...arguments);
        this.version = 0;
        this.encodes = exports.nullPlaceholder;
        this.name = (0, utils_1.getBuiltInEncodingName)("null");
        this.fixedIndex = fixed_indexes_1.FixedIndexes.NullProto;
        this.encode = getPrototypeEncoder(null);
        this.decoder = getPrototypeDecoder(null);
    }
});
function getPrototypeDecoder(encodes) {
    return {
        init: exports.objectEncoding.decoder.init,
        create(encodedValue, ctx) {
            return Object.create(encodes);
        }
    };
}
exports.getPrototypeDecoder = getPrototypeDecoder;
function getPrototypeEncoder(proto) {
    return (input, ctx) => {
        const result = encodeObject(input, ctx, false);
        ctx._isImplicit = false;
        return result;
    };
}
exports.getPrototypeEncoder = getPrototypeEncoder;
class UnsupportedEncoding extends interface_1.PrototypeEncoding {
    constructor(name, encodes, fixedIndex) {
        super();
        this.name = name;
        this.encodes = encodes;
        this.fixedIndex = fixedIndex;
        this.version = 0;
        this.decoder = {
            create(encodedValue, ctx) {
                return new interface_1.PreszrUnsupportedValue(ctx.self.name.slice(1));
            }
        };
        this.encodes = encodes;
        this.name = name;
    }
    encode(input, ctx) {
        return 0;
    }
}
exports.unsupportedEncodings = [
    [Function, fixed_indexes_1.FixedIndexes.Function],
    [opt_types_1._GeneratorFunction, fixed_indexes_1.FixedIndexes.GeneratorFunction],
    [opt_types_1._Generator, fixed_indexes_1.FixedIndexes.Generator],
    [Promise, fixed_indexes_1.FixedIndexes.Promise],
    [WeakSet, fixed_indexes_1.FixedIndexes.WeakSet],
    [WeakMap, fixed_indexes_1.FixedIndexes.WeakMap],
    [opt_types_1._AsyncGenerator, fixed_indexes_1.FixedIndexes.AsyncGenerator],
    [opt_types_1._AsyncGeneratorFunction, fixed_indexes_1.FixedIndexes.AsyncGeneratorFunction],
    [opt_types_1._FinalizationRegistry, fixed_indexes_1.FixedIndexes.FinalizationRegistry],
    [opt_types_1._AsyncFunction, fixed_indexes_1.FixedIndexes.AsyncFunction],
    [opt_types_1._WeakRef, fixed_indexes_1.FixedIndexes.WeakRef]
]
    .filter(x => x[0])
    .map(([ctor, index]) => {
    const name = (0, utils_1.getBuiltInEncodingName)((0, utils_1.getPrototypeName)(ctor.prototype));
    return new UnsupportedEncoding(name, ctor.prototype, index);
});
exports.unsupportedEncodings.push(...[
    [opt_types_1._SetIteratorProto, fixed_indexes_1.FixedIndexes.SetIterator],
    [opt_types_1._MapIteratorProto, fixed_indexes_1.FixedIndexes.MapIterator],
    [opt_types_1._ArrayIteratorProto, fixed_indexes_1.FixedIndexes.ArrayIterator]
].map(([proto, index]) => new UnsupportedEncoding((0, utils_1.getBuiltInEncodingName)((0, utils_1.getPrototypeName)(proto)), proto, index)));
//# sourceMappingURL=objects.js.map