import { EncodedEntity, ScalarValue } from "./data";
/**
 * The context used by the encoding process.
 */
export interface EncodeContext {
    readonly self: PrototypeEncoding;
    /**
     * Encodes the given input. For entities, it will recursively encode them, add
     * them to the final output as a side effect, and return a reference. For other
     * values, it will return them as-is or encode them, usually as a string.
     * @param value
     */
    encode(value: any): ScalarValue;
    /**
     * For internal use only.
     */
    _isImplicit: boolean;
    /**
     * Sets the metadata for this entity. The metadata can be any JSON-legal value,
     * including an object. It doesn't do anything, but can be accessed while decoding.
     */
    metadata: any;
}
/**
 * The context used by the create stage of the decoding process. Only
 * exposes the entity's metadata.
 */
export interface CreateContext {
    self: PrototypeEncoding;
    metadata: any;
}
/**
 * The context used by the init stage of the decoding process. Allows
 * resolving references to other entities.
 */
export interface InitContext {
    metadata: any;
    self: PrototypeEncoding;
    decode(value: ScalarValue): unknown;
}
/**
 * The decoding logic for a prototype encoding.
 */
export interface Decoder {
    create(encoded: EncodedEntity, ctx: CreateContext): unknown;
    init?(target: unknown, encoded: EncodedEntity, ctx: InitContext): void;
}
/**
 * Specifies a prototype encoding. Missing fields will be filled in automatically.
 */
export interface PrototypeSpecifier {
    name?: string | null;
    version?: number;
    encodes: object | Function;
    decoder?: Decoder;
    encode?(input: any, ctx: EncodeContext): EncodedEntity;
}
/**
 * A full symbol encoding.
 */
export interface SymbolSpecifier {
    name: string;
    encodes: symbol;
}
/**
 * A full preszr encoding of any type.
 */
export declare type Encoding = SymbolEncoding | PrototypeEncoding<object>;
/**
 * These encodings specifiers can't be confused for configuration objects or
 * prototypes.
 */
export declare type BasicSpecifier = symbol | Function;
/**
 * An encoding specifier. Can be a symbol or constructor for a shorthand
 * symbol or prototype encoding. You can also give symbol or prototype encoding specifier
 * if you want to be more explicit.
 */
export declare type EncodingSpecifier = BasicSpecifier | PrototypeSpecifier | SymbolSpecifier;
/**
 * Configuration for an Preszr instance.
 */
export interface PreszrConfig {
    /**
     * An array of encoding specifiers. If you put your constructors and symbols here,
     * the Preszr will recognize them.
     */
    encodes: EncodingSpecifier[];
}
export declare abstract class BaseEncoding {
    abstract readonly name: string;
    abstract readonly fixedIndex?: number;
    abstract readonly key: string;
    protected abstract _toString(): string;
    get [Symbol.toStringTag](): string;
    toString(): string;
}
export declare class SymbolEncoding extends BaseEncoding {
    readonly name: string;
    readonly encodes: symbol;
    readonly fixedIndex?: number | undefined;
    readonly metadata?: any;
    constructor(name: string, encodes: symbol, fixedIndex?: number | undefined, metadata?: any);
    protected _toString(): string;
    get key(): string;
    static fromSymbol(s: symbol): SymbolEncoding;
    static fromSpecifier(spec: SymbolSpecifier): SymbolEncoding;
}
export declare class PreszrUnsupportedValue {
    readonly type: string;
    constructor(type: string);
}
export declare abstract class PrototypeEncoding<T extends {
    constructor: Function;
} = object> extends BaseEncoding {
    abstract readonly name: string;
    abstract readonly version: number;
    abstract readonly encodes: T;
    abstract readonly fixedIndex?: number;
    abstract encode(x: T, ctx: EncodeContext): EncodedEntity;
    abstract decoder: Decoder;
    get key(): string;
    protected _toString(): string;
    mustBeCompatible(other: PrototypeEncoding): void;
}
//# sourceMappingURL=interface.d.ts.map